/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QAccelerometer>
#include <QAccelerometerReading>
#include <QByteArray>
#include <QColor>
#include <QFrame>
#include <QFont>
#include <QFontMetrics>
#include <QFontMetrics>
#include <QGridLayout>
#include <QGuiApplication>
#include <QHBoxLayout>
#include <QIcon>
#include <QLabel>
#include <QProgressBar>
#include <QResizeEvent>
#include <QScreen>
#include <QSize>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QToolButton>
#include <QVBoxLayout>
#include <QWidget>
#include <cmath>

#ifndef Q_OS_ANDROID
#define OFFLINE_TEST_MODE   // simulates data for offline testing purposes
#endif

// #define ENABLE_TEST_MODE    // adds additional test and display options

#ifdef OFFLINE_TEST_MODE
#include <QTimer>
#include <QTime>
#include <stdlib.h>
#include <cstdlib>
#include <time.h>
#endif

#ifdef ENABLE_TEST_MODE
#include <QElapsedTimer>
#include <QList>
#include <QPair>
#include <QFile>

	QLabel								*g_debug_label_1;
	QLabel								*g_debug_label_2;
	QLabel								*g_debug_label_3;
	QLabel								*g_debug_label_4;
	QLabel								*g_debug_label_5;

#endif

#include "button.h"
#include "../core/graphwidget.h"

#include "vibrationmeter.h"

TVibrationMeter::TVibrationMeter(
	const QWidget *parent)
: QMainWindow(const_cast<QWidget*>(parent))
{
    QWidget                             *central_widget;
	QScreen								*screen;
	QFont								copyright_font;
	QIcon								button_icons_grey_green;
	QIcon								button_icon_yellow;
    QSizePolicy                         size_policy_minimum_minimum(QSizePolicy::Minimum, QSizePolicy::Minimum);
	QFontMetrics						font_metrics(QApplication::font());
    int									button_dimension(font_metrics.height() * 8);

    size_policy_minimum_minimum.setHorizontalStretch(0);
    size_policy_minimum_minimum.setVerticalStretch(0);

	this->resize(1200,800);
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
    copyright_font.setPointSize(8);     // original 8
	copyright_font.setItalic(true);
	
    central_widget = new QWidget(this);
    this->setCentralWidget(central_widget);

    d_widget_layout = new QGridLayout(central_widget);

    d_acceleration_graph_widget = new TGraphWidget(central_widget);
	d_acceleration_graph_widget->Enable_Direct_Render(true);
	d_acceleration_graph_widget->EnableMouseCoordinateDisplay(false);
	d_acceleration_graph_widget->setFont(QApplication::font());
    d_widget_layout->addWidget(d_acceleration_graph_widget,0,0,1,1);

    d_fourier_graph_widget = new TGraphWidget(central_widget);
	d_fourier_graph_widget->Enable_Direct_Render(true);
	d_fourier_graph_widget->SetHLegendMode(TGraphWidget::INTEGER_LEGEND_PRECISION);
    d_fourier_graph_widget->EnableMouseCoordinateDisplay(false);
	d_fourier_graph_widget->EnableYIncludesZero(true);
	d_fourier_graph_widget->SetScaleMinimumY(1.0,false);
	d_fourier_graph_widget->setFont(QApplication::font());
	d_widget_layout->addWidget(d_fourier_graph_widget,1,0,1,1);
	
	d_control_layout = new QGridLayout();
	
	d_result_x_label = new QLabel(central_widget);
	d_result_x_label->setAlignment(Qt::AlignLeft|Qt::AlignVCenter);
    d_result_x_label->setFont(QApplication::font());
	d_control_layout->addWidget(d_result_x_label,0,0,1,2);
	
	d_result_y_label = new QLabel(central_widget);
	d_result_y_label->setAlignment(Qt::AlignLeft|Qt::AlignVCenter);
    d_result_y_label->setFont(QApplication::font());
    d_control_layout->addWidget(d_result_y_label,1,0,1,2);
	
	d_result_z_label = new QLabel(central_widget);
	d_result_z_label->setAlignment(Qt::AlignLeft|Qt::AlignVCenter);
    d_result_z_label->setFont(QApplication::font());
    d_control_layout->addWidget(d_result_z_label,2,0,1,2);
	
	d_spacer_item = new QSpacerItem(0, 0, QSizePolicy::Minimum, QSizePolicy::MinimumExpanding);
	d_control_layout->addItem(d_spacer_item,3,0,1,2);

	d_enable_x_button = new TButton(central_widget);
	d_enable_x_button->setCheckable(true);
	d_enable_x_button->setChecked(true);
    d_enable_x_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_enable_x_button->setSizePolicy(size_policy_minimum_minimum);
    d_enable_x_button->setFont(QApplication::font());
    d_control_layout->addWidget(d_enable_x_button,4,0,1,1);
	
	d_enable_y_button = new TButton(central_widget);
	d_enable_y_button->setCheckable(true);
	d_enable_y_button->setChecked(true);
    d_enable_y_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_enable_y_button->setSizePolicy(size_policy_minimum_minimum);
    d_enable_y_button->setFont(QApplication::font());
    d_control_layout->addWidget(d_enable_y_button,5,0,1,1);
	
	d_enable_z_button = new TButton(central_widget);
	d_enable_z_button->setCheckable(true);
	d_enable_z_button->setChecked(true);
    d_enable_z_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_enable_z_button->setSizePolicy(size_policy_minimum_minimum);
    d_enable_z_button->setFont(QApplication::font());
    d_control_layout->addWidget(d_enable_z_button,6,0,1,1);
	
	d_live_button = new TButton(central_widget);
	d_live_button->setCheckable(true);
	d_live_button->setChecked(true);
    d_live_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_live_button->setSizePolicy(size_policy_minimum_minimum);
    d_live_button->setFont(QApplication::font());
    d_control_layout->addWidget(d_live_button,4,1,1,1);
	
	d_zero_button = new TButton(central_widget);
	d_zero_button->setCheckable(true);
	d_zero_button->setChecked(true);
    d_zero_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_zero_button->setSizePolicy(size_policy_minimum_minimum);
    d_zero_button->setFont(QApplication::font());
    d_control_layout->addWidget(d_zero_button,5,1,1,1);
	
	d_stop_button = new TButton(central_widget);
	d_stop_button->setCheckable(true);
    d_stop_button->setMinimumSize(QSize(button_dimension,button_dimension));
    d_stop_button->setSizePolicy(size_policy_minimum_minimum);
	d_stop_button->setFont(QApplication::font());
	d_control_layout->addWidget(d_stop_button,6,1,1,1);
	
	d_live_update_progress = new QProgressBar(central_widget);
	d_live_update_progress->setEnabled(false);
	d_live_update_progress->setFont(QApplication::font());
	d_control_layout->addWidget(d_live_update_progress,7,0,1,2);
	
	d_widget_layout->addLayout(d_control_layout,0,1,2,1);	// assume wide startup

	d_copyright_label = new QLabel(this);
	d_copyright_label->setFont(copyright_font);
	d_widget_layout->addWidget(d_copyright_label,2,0,1,2);	// assuming wide startup
	
    d_acceleration_graph_widget->Title("Acceleration");
    d_acceleration_graph_widget->VerticalLegend("mm/sec^2");
    d_acceleration_graph_widget->HorizontalLegend("Seconds");

    d_fourier_graph_widget->Title("Frequency Spectrum");
    d_fourier_graph_widget->VerticalLegend("Relative Amplitude");
    d_fourier_graph_widget->HorizontalLegend("Frequency HZ");

    d_accel = new QAccelerometer(this);
	
#ifdef ENABLE_TEST_MODE
	qrangelist							range_list;
	qrangelist::const_iterator			range_iter;
	QString								text;

	// create debug label lines
	g_debug_label_1 = new QLabel(this);
	g_debug_label_1->setWordWrap(true);
	d_widget_layout->addWidget(g_debug_label_1,3,0,1,2);	// assuming wide startup

	g_debug_label_2 = new QLabel(this);
	g_debug_label_2->setWordWrap(true);
	d_widget_layout->addWidget(g_debug_label_2,4,0,1,2);	// assuming wide startup

	g_debug_label_3 = new QLabel(this);
	g_debug_label_3->setWordWrap(true);
	d_widget_layout->addWidget(g_debug_label_3,5,0,1,2);	// assuming wide startup

	g_debug_label_4 = new QLabel(this);
	g_debug_label_4->setWordWrap(true);
	d_widget_layout->addWidget(g_debug_label_4,6,0,1,2);	// assuming wide startup

	g_debug_label_5 = new QLabel(this);
	g_debug_label_5->setWordWrap(true);
	d_widget_layout->addWidget(g_debug_label_5,7,0,1,2);	// assuming wide startup
	
	// display supported sample rates
	range_list = d_accel->availableDataRates();
	text = QStringLiteral("supported data rates:\n");
	
	for(range_iter = range_list.begin();range_iter != range_list.end();++range_iter)
	{
		text += QString("%1 - %2/n").arg((*range_iter).first).arg((*range_iter).second);
	}
	g_debug_label_1->setText(text);
	
#endif
	
	// defaults;
	d_frame_size = 512;
	d_frame_count = 0;
	d_display_mode = TVibrationMeter::DISPLAY_WIDE;
	d_initialize = true;
	d_live_display = true;
	d_enable_x = true;
	d_enable_y = true;
	d_enable_z = true;
	d_enable_filter = false;
	d_enable_zero = true;
	d_stop_updates = false;
	d_update_display = 0;
	d_stop_button->setChecked(true);
	d_live_update_progress->setRange(0,d_frame_size);
	
    this->setWindowTitle(QStringLiteral("Vibration Meter - 2.0"));
	
    d_enable_x_button->setText(QStringLiteral("X"));
    d_enable_y_button->setText(QStringLiteral("Y"));
    d_enable_z_button->setText(QStringLiteral("Z"));
    d_zero_button->setText(QStringLiteral("Zero"));
	d_live_button->setText(QStringLiteral("Live"));
	d_stop_button->setText(QStringLiteral("Stop"));
    d_copyright_label->setText(QStringLiteral("Copyright (C) 2022 Select Calibration Incorporated.  www.selectcalibration.ca"));
	
	screen = QGuiApplication::primaryScreen();
	screen->setOrientationUpdateMask(Qt::PortraitOrientation | Qt::LandscapeOrientation | Qt::InvertedPortraitOrientation | Qt::InvertedLandscapeOrientation);
	
	connect(screen,&QScreen::orientationChanged,this,&TVibrationMeter::Screen_Orientation_Changed);
    connect(d_accel,&QAccelerometer::readingChanged,this,&TVibrationMeter::Update_Accelerometer);
	connect(d_enable_x_button,&TButton::toggled,this,&TVibrationMeter::Enable_X_Toggled);
	connect(d_enable_y_button,&TButton::toggled,this,&TVibrationMeter::Enable_Y_Toggled);
	connect(d_enable_z_button,&TButton::toggled,this,&TVibrationMeter::Enable_Z_Toggled);
	connect(d_zero_button,&TButton::toggled,this,&TVibrationMeter::Enable_Zero_Toggled);
	connect(d_live_button,&TButton::toggled,this,&TVibrationMeter::Live_Toggled);
	connect(d_stop_button,&TButton::toggled,this,&TVibrationMeter::Stop_Toggled);
	

#ifdef OFFLINE_TEST_MODE
		
	srand (time(NULL));
	
	QTimer *timer = new QTimer();
	
	timer->setTimerType(Qt::PreciseTimer);
	connect(timer,&QTimer::timeout,this,&TVibrationMeter::Update_Accelerometer);
	
    timer->start(10);	// 100 hz
	
#endif
	
	d_accel->setActive(true);
    d_accel->start();
}

TVibrationMeter::~TVibrationMeter()
{

}

void TVibrationMeter::Screen_Orientation_Changed(
	Qt::ScreenOrientation				orientation)
{
	qWarning("orientation changed");
	
	if(orientation == Qt::LandscapeOrientation ||
	   orientation == Qt::InvertedLandscapeOrientation)
	{
		if(d_display_mode != TVibrationMeter::DISPLAY_WIDE)
		{
			d_display_mode = TVibrationMeter::DISPLAY_WIDE;
			this->Update_Layout();
		}
	}
	else if(orientation == Qt::PortraitOrientation ||
			orientation == Qt::InvertedPortraitOrientation)
	{
		if(d_display_mode != TVibrationMeter::DISPLAY_TALL)
		{
			d_display_mode = TVibrationMeter::DISPLAY_TALL;
			this->Update_Layout();
		}
	}
}

void TVibrationMeter::Update_Accelerometer(void)
{  
#ifdef ENABLE_TEST_MODE
    QElapsedTimer						elapsed_timer;
	static bool							enable_data_dump(false);
	
    elapsed_timer.start();
#endif

    qint64								timestamp;
    TVector2							data;
    double								mean_accel;
	double								mean_accel_weight;
	double								dval;
    int									cntr;
    std::vector<TVector2>::iterator		iter;
    static const int					average_sample_size(32);
	
#ifdef OFFLINE_TEST_MODE
	QTime								time(QTime::currentTime());
	double								angle_increment;
	static bool							update_offsets(false);
	static double						offset_x(0.0);
	static double						offset_y(0.0);
	static double						offset_z(0.0);
	static double						current_angle(0);
	
	// ignore timer signal when running test signals
	if(d_stop_updates)
	{
		return;
	}

#else
	QAccelerometerReading				*accel_reading;
	
	accel_reading = d_accel->reading();
	
	if(!accel_reading)
	{
		return;
	}
	
#endif
	
    if(d_initialize)
    {
		d_acceleration_data_x.clear();
		d_acceleration_data_y.clear();
		d_acceleration_data_z.clear();
		
#ifdef OFFLINE_TEST_MODE
		d_start_timestamp = static_cast<qint64>(time.msecsSinceStartOfDay()) * 1000;
		d_draw_timestamp = d_start_timestamp;
#else
		d_start_timestamp = accel_reading->timestamp();
		d_draw_timestamp = d_start_timestamp;
#endif

#ifdef OFFLINE_TEST_MODE
		d_mean_ax = 0.2 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		d_mean_ay = 0.2 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		d_mean_az = 0.2 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		
		offset_x = 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		offset_y = 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		offset_z = 9.8 + 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));

        current_angle = 0;
        angle_increment = 0;
#else
		d_mean_ax = accel_reading->x();
		d_mean_ay = accel_reading->y();
		d_mean_az = accel_reading->z();
#endif

        d_initialize = false;
		d_update_display = 0;
		
		d_frame_count = 0;
		d_live_update_progress->setValue(d_frame_count);
    }
    else
    {
#ifdef OFFLINE_TEST_MODE
		timestamp = static_cast<qint64>(time.msecsSinceStartOfDay()) * 1000;
#else
		timestamp = accel_reading->timestamp();
#endif
		
		if((timestamp - d_draw_timestamp) > 250000)	// redraw 4 times a second.  minimum redraw time may be as high as 50 ms
        {
            d_update_display++;
            d_draw_timestamp = timestamp;
        }

        data.x = static_cast<double>(timestamp - d_start_timestamp) * 0.000001;

#ifdef OFFLINE_TEST_MODE
		
		// update current angle.  1 hz
		current_angle = data.x * 6.28318530717959;
		
		if(current_angle > 6.28318530717959)
		{
			current_angle -= 6.28318530717959;
		}

		
//		data.y = 0.1 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		data.y = 0;
		
		// add harmonics
		data.y += (sin(current_angle * 1));
		data.y += (sin(current_angle * 3));
		data.y += (sin(current_angle * 5));
		
		// add long period square wave
		// if((time.second() / 15) % 2)
		if(time.minute() % 2)
		{
			data.y += offset_x;
			update_offsets = true;
		}
		else
		{
			if(update_offsets)
			{
				offset_x = 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
				offset_y = 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
				offset_z = 9.8 + 10.0 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
				
//				qWarning(QString("offset x,y,z: %1 %2 %3").arg(offset_x,6,'f',3).arg(offset_y,6,'f',3).arg(offset_z,6,'f',3).toLatin1());
				
				update_offsets = false;
			}
		}

#else
		data.y = accel_reading->x() * 1000.0;	// mm/sec
#endif

 //       g_debug_label_3->setText(QString("x data: %1 x mean offset: %2").arg(data.y,0,'f',4).arg(d_mean_ax,0,'f',4));

		data.y -= d_mean_ax;
		
		d_acceleration_data_x.push_back(data);
		

#ifdef OFFLINE_TEST_MODE
		
//		data.y = 0.1 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		data.y = 0;
		
		// add harmonics
		data.y += (2 * sin(current_angle * 7));
		data.y += (2 * sin(current_angle * 9));
		data.y += (2 * sin(current_angle * 11));
		
		// add long period square wave
		// if((time.second() / 15) % 2)
		if(time.minute() % 2)
		{
			data.y += offset_y;
		}

#else
		data.y = accel_reading->y() * 1000.0;	// mm/sec;
#endif
		data.y -= d_mean_ay;

		d_acceleration_data_y.push_back(data);

#ifdef OFFLINE_TEST_MODE
		
//		data.y = 0.1 * (-0.5 + static_cast<double>(rand()) / static_cast<double>(RAND_MAX));
		data.y = -29.8;
		
		// add harmonics
		data.y += (3 * sin(current_angle * 13));
		data.y += (3 * sin(current_angle * 15));
		data.y += (3 * sin(current_angle * 17));
		
		// add long period square wave
		// if((time.second() / 15) % 2)
		if(time.minute() % 2)
		{
			data.y += offset_z;
		}
#else
		data.y = accel_reading->z() * 1000.0;	// mm/sec;
#endif
		data.y -= d_mean_az;

		d_acceleration_data_z.push_back(data);
		
#ifdef ENABLE_TEST_MODE
		g_debug_label_4->setText(QString("Completion time prior to mean filter: %1 ms").arg(elapsed_timer.elapsed()));
		// g_debug_label_1->setText(QString("X data: %1 Y data: %2  Z data: %3").arg(accel_reading->x(),8,'f',4).arg(accel_reading->y(),8,'f',4).arg(accel_reading->z(),8,'f',4));
		
#endif

		if(!d_live_display)
		{
			d_frame_count++;
			d_live_update_progress->setValue(d_frame_count);
		}

        if(d_enable_zero)
        {
			// mean value calculated as a weighted 'triangle'
			
			// x axis bias
			mean_accel = 0.0;
			mean_accel_weight = 0.0;
			for(iter = (d_acceleration_data_x.end()-1),cntr = 0;!(iter < d_acceleration_data_x.begin());--iter,++cntr)
			{
				dval = static_cast<double>(average_sample_size - cntr) / static_cast<double>(average_sample_size);
				
 //               g_debug_label_4->setText(QString("average_sample_size: %1 cntr: %2  dval: %3").arg(average_sample_size).arg(cntr).arg(dval,0,'f',6));

				mean_accel_weight += dval;
                mean_accel += (dval * (*iter).y);

 //               g_debug_label_4->setText(QString("weight: %1 data: %2").arg(dval,0,'f',6).arg((*iter).y,0,'f',6));

				if(!(cntr < average_sample_size))
				{
					break;
				}
			}

			if(cntr)
			{
//                g_debug_label_5->setText(QString("cntr: %1  mean_accel: %2  mean_accel_weight: %3").arg(cntr,0,'f',6).arg(mean_accel,0,'f',6).arg(mean_accel_weight,0,'f',6));

				dval = (mean_accel / mean_accel_weight);

//                g_debug_label_4->setText(QString("d_mean_ax: %1 d_mean_ax offset: %2").arg(d_mean_ax,0,'f',4).arg(dval,0,'f',4));

				d_mean_ax += (dval);
				
				
				// offset existing data by change
				for(iter = d_acceleration_data_x.end() - 1,cntr = 0;!(iter < d_acceleration_data_x.begin());--iter,++cntr)
				{
					(*iter).y -= dval;
					
					if(!(cntr < average_sample_size))
					{
						break;
					}
				}
			}

			// y axis bias
			mean_accel = 0.0;
			mean_accel_weight = 0.0;
			for(iter = (d_acceleration_data_y.end()-1),cntr = 0;!(iter < d_acceleration_data_y.begin());--iter,++cntr)
			{
				dval = static_cast<double>(average_sample_size - cntr) / static_cast<double>(average_sample_size);
				
				mean_accel_weight += dval;
				mean_accel += (dval * (*iter).y);
				
				if(!(cntr < average_sample_size))
				{
					break;
				}
			}
			
			if(cntr)
			{
				dval = (mean_accel / mean_accel_weight);
				d_mean_ay += (dval);
				
				
				// offset existing data by change
				for(iter = d_acceleration_data_y.end() - 1,cntr = 0;!(iter < d_acceleration_data_y.begin());--iter,++cntr)
				{
					(*iter).y -= dval;
					
					if(!(cntr < average_sample_size))
					{
						break;
					}
				}
			}

			// z axis bias
			mean_accel = 0.0;
			mean_accel_weight = 0.0;
			for(iter = (d_acceleration_data_z.end()-1),cntr = 0;!(iter < d_acceleration_data_z.begin());--iter,++cntr)
			{
				dval = static_cast<double>(average_sample_size - cntr) / static_cast<double>(average_sample_size);
				
				mean_accel_weight += dval;
				mean_accel += (dval * (*iter).y);
				
				if(!(cntr < average_sample_size))
				{
					break;
				}
			}
			
			if(cntr)
			{
				dval = (mean_accel / mean_accel_weight);
				d_mean_az += (dval);
				
				// offset existing data by change
				for(iter = d_acceleration_data_z.end() - 1,cntr = 0;!(iter < d_acceleration_data_z.begin());--iter,++cntr)
				{
					(*iter).y -= dval;
					
					if(!(cntr < average_sample_size))
					{
						break;
					}
				}
			}
        }
		else
		{
			d_mean_ax = 0.0;
			d_mean_ay = 0.0;
			d_mean_az = 0.0;
		}


        if(d_enable_x)
        {
            while(static_cast<int>(d_acceleration_data_x.size()) > d_frame_size)
			{
				d_acceleration_data_x.erase(d_acceleration_data_x.begin());
			}
        }

        if(d_enable_y)
        {
            while(static_cast<int>(d_acceleration_data_y.size()) > d_frame_size)
			{
				d_acceleration_data_y.erase(d_acceleration_data_y.begin());
			}
        }

        if(d_enable_z)
        {
            while(static_cast<int>(d_acceleration_data_z.size()) > d_frame_size)
			{
				d_acceleration_data_z.erase(d_acceleration_data_z.begin());
			}
        }

		if(d_live_display)
		{
			// updates to display are alternated to reducing processing time
			switch(d_update_display)
			{
				case 1:
					this->Update_Acceleration_View();
					d_update_display = 2;
					break;
					
				case 2:
					this->Update_Fourier_View();
					d_update_display = 3;
					break;
					
				case 3:
					this->Update_Analysis_Data();
					d_update_display = 0;
					break;
					
				default:
					d_update_display = 0;
					break;
			}
		}
		else
		{
			if(d_frame_count == d_frame_size)
			{
				d_acceleration_data_last_frame_x = d_acceleration_data_x;
				d_acceleration_data_last_frame_y = d_acceleration_data_y;
				d_acceleration_data_last_frame_z = d_acceleration_data_z;

				this->Update_Acceleration_View();
				this->Update_Fourier_View();
				this->Update_Analysis_Data();

				d_initialize = true;
			}
		}
		
#ifdef ENABLE_TEST_MODE
		g_debug_label_2->setText(QString("Completion time with display refresh: %1 ms").arg(elapsed_timer.elapsed()));
#endif
    }
	
#ifdef ENABLE_TEST_MODE
	
	if(enable_data_dump && (static_cast<int>(d_acceleration_data_x.size()) == d_frame_size))
	{
		QFile file("/Users/Ron/output.dat");
		
		if(file.open(QIODevice::WriteOnly | QIODevice::Text))
		{
			qWarning("writing test file ...");
			for(iter = d_acceleration_data_x.begin();iter != d_acceleration_data_x.end();++iter)
			{
				file.write(QString("%1,%2\n").arg((*iter).x,0,'f',6).arg((*iter).y,0,'f',6).toLatin1());
			}

			file.close();
			qWarning("writing test file done");

		}
		
		enable_data_dump = false;
	}
#endif
}

void TVibrationMeter::Enable_X_Toggled(
	bool								state)
{
	d_enable_x = state;
	
	if(d_stop_updates)
	{
		this->Update_Acceleration_View();
		this->Update_Fourier_View();
		this->Update_Analysis_Data();
	}
}

void TVibrationMeter::Enable_Y_Toggled(
	bool								state)
{
	d_enable_y = state;
	
	if(d_stop_updates)
	{
		this->Update_Acceleration_View();
		this->Update_Fourier_View();
		this->Update_Analysis_Data();
	}
}

void TVibrationMeter::Enable_Z_Toggled(
	bool								state)
{
	d_enable_z = state;
	
	if(d_stop_updates)
	{
		this->Update_Acceleration_View();
		this->Update_Fourier_View();
		this->Update_Analysis_Data();
	}
}

void TVibrationMeter::Enable_Zero_Toggled(
	bool								state)
{
	d_enable_zero = state;
}

void TVibrationMeter::Live_Toggled(
	bool								state)
{
	if(state)
	{
		d_live_display = true;
//		d_initialize = true;
		d_live_update_progress->setValue(0);
		d_live_update_progress->setEnabled(false);
		
		d_enable_x_button->setEnabled(true);
		d_enable_y_button->setEnabled(true);
		d_enable_z_button->setEnabled(true);
	}
	else
	{
		d_live_display = false;
		d_initialize = true;
		d_live_update_progress->setEnabled(true);
		
		d_enable_x_button->setEnabled(false);
		d_enable_y_button->setEnabled(false);
		d_enable_z_button->setEnabled(false);
	}
}

void TVibrationMeter::Stop_Toggled(
	bool								state)
{
	if(state)
	{
		d_stop_updates = false;
		d_stop_button->setText(QStringLiteral("Stop"));
		
		// must reset data if stopped
		d_initialize = true;

		d_accel->setActive(true);
		d_accel->start();
		
		d_live_button->setEnabled(true);
		d_zero_button->setEnabled(true);
		
		d_enable_x_button->setEnabled(d_live_display);
		d_enable_y_button->setEnabled(d_live_display);
		d_enable_z_button->setEnabled(d_live_display);
	}
	else
	{
		d_stop_updates = true;
		d_stop_button->setText(QStringLiteral("Start"));
	
		d_accel->stop();
		d_accel->setActive(false);
		
		d_live_button->setEnabled(false);
		d_zero_button->setEnabled(false);
		
		d_enable_x_button->setEnabled(true);
		d_enable_y_button->setEnabled(true);
		d_enable_z_button->setEnabled(true);
		
		if(!d_live_display)
		{
			this->Update_Acceleration_View();
			this->Update_Fourier_View();
			this->Update_Analysis_Data();
			
			d_live_update_progress->setValue(0);
		}
	}
}

void TVibrationMeter::resizeEvent(
	QResizeEvent						*event)
{
	QSize								size(event->size());
	
	QMainWindow::resizeEvent(event);
	
	if(size.width() > 0 && size.height() > 0)
	{
		if(size.width() > size.height())
		{
			if(d_display_mode != TVibrationMeter::DISPLAY_WIDE)
			{
				d_display_mode = TVibrationMeter::DISPLAY_WIDE;
				this->Update_Layout();
			}
		}
		else
		{
			if(d_display_mode != TVibrationMeter::DISPLAY_TALL)
			{
				d_display_mode = TVibrationMeter::DISPLAY_TALL;
				this->Update_Layout();
			}
		}
	}
	
}

void TVibrationMeter::Update_Layout(void)
{
	d_widget_layout->removeItem(d_control_layout);
	d_widget_layout->removeWidget(d_copyright_label);

	d_control_layout->removeWidget(d_result_x_label);
	d_control_layout->removeWidget(d_result_y_label);
	d_control_layout->removeWidget(d_result_z_label);
	d_control_layout->removeItem(d_spacer_item);
	d_control_layout->removeWidget(d_enable_x_button);
	d_control_layout->removeWidget(d_enable_y_button);
	d_control_layout->removeWidget(d_enable_z_button);
	d_control_layout->removeWidget(d_live_button);
	d_control_layout->removeWidget(d_zero_button);
	d_control_layout->removeWidget(d_live_update_progress);
	d_control_layout->removeWidget(d_stop_button);
	
	if(d_display_mode == TVibrationMeter::DISPLAY_WIDE)
	{
		d_spacer_item->changeSize(0,0,QSizePolicy::Minimum,QSizePolicy::MinimumExpanding);
		
		d_widget_layout->addLayout(d_control_layout,0,1,2,1);
		d_widget_layout->addWidget(d_copyright_label,2,0,1,2);
		
		d_control_layout->addWidget(d_result_x_label,0,0,1,2);
		d_control_layout->addWidget(d_result_y_label,1,0,1,2);
		d_control_layout->addWidget(d_result_z_label,2,0,1,2);
		d_control_layout->addItem(d_spacer_item,3,0,1,2);
		d_control_layout->addWidget(d_enable_x_button,4,0,1,1);
		d_control_layout->addWidget(d_enable_y_button,5,0,1,1);
		d_control_layout->addWidget(d_enable_z_button,6,0,1,1);
		d_control_layout->addWidget(d_live_button,4,1,1,1);
		d_control_layout->addWidget(d_zero_button,5,1,1,1);
		d_control_layout->addWidget(d_stop_button,6,1,1,1);
		d_control_layout->addWidget(d_live_update_progress,7,0,1,2);
	}
	else // if(d_display_mode == TVibrationMeter::DISPLAY_TALL)
	{
		d_spacer_item->changeSize(0,0,QSizePolicy::MinimumExpanding,QSizePolicy::Minimum);

		d_widget_layout->addLayout(d_control_layout,2,0,1,1);
		d_widget_layout->addWidget(d_copyright_label,4,0,1,1);
		
		d_control_layout->addWidget(d_result_x_label,0,0,1,1);
		d_control_layout->addWidget(d_result_y_label,1,0,1,1);
		d_control_layout->addWidget(d_result_z_label,2,0,1,1);
		d_control_layout->addItem(d_spacer_item,0,1,3,1);
		d_control_layout->addWidget(d_enable_x_button,0,2,1,1);
		d_control_layout->addWidget(d_enable_y_button,1,2,1,1);
		d_control_layout->addWidget(d_enable_z_button,2,2,1,1);
		d_control_layout->addWidget(d_live_button,0,3,1,1);
		d_control_layout->addWidget(d_zero_button,1,3,1,1);
		d_control_layout->addWidget(d_stop_button,2,3,1,1);
		d_control_layout->addWidget(d_live_update_progress,3,0,1,4);
	}
}

void TVibrationMeter::Update_Acceleration_View(void)
{
	int									set;

	d_acceleration_graph_widget->DeleteDataSets();
	
	if(!d_live_display)
	{
		if(d_enable_x)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(212,0,0));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_last_frame_x);
		}
		
		if(d_enable_y)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(0,212,0));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_last_frame_y);
		}
		
		if(d_enable_z)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(0,0,212));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_last_frame_z);
		}
	}
	else
	{

		if(d_enable_x)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(212,0,0));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_x);
		}
		
		if(d_enable_y)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(0,212,0));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_y);
		}
		
		if(d_enable_z)
		{
			set = d_acceleration_graph_widget->CreateDataSet(QColor(0,0,212));
			d_acceleration_graph_widget->AddPoints(set,d_acceleration_data_z);
		}
	}
	
	d_acceleration_graph_widget->UpdateGraph();
}

void TVibrationMeter::Update_Fourier_View(void)
{
	std::vector<TVector2>				fourier_data_x;
	std::vector<TVector2>				fourier_data_y;
	std::vector<TVector2>				fourier_data_z;
	int									set;
	double								fft_max;
	double								fft_max_x(0.0);
	double								fft_max_y(0.0);
	double								fft_max_z(0.0);
	static const double					ZERO_EPSILON(0.000001);

	d_fourier_graph_widget->DeleteDataSets();
	
	if(!d_live_display)
	{
		fourier_data_x = d_fourier_transform.Process(d_acceleration_data_last_frame_x);
		fft_max_x = this->Get_Max_Value(fourier_data_x);

		fourier_data_y = d_fourier_transform.Process(d_acceleration_data_last_frame_y);
		fft_max_y = this->Get_Max_Value(fourier_data_y);

		fourier_data_z = d_fourier_transform.Process(d_acceleration_data_last_frame_z);
		fft_max_z = this->Get_Max_Value(fourier_data_z);

	}
	else
	{
		fourier_data_x = d_fourier_transform.Process(d_acceleration_data_x);
		fft_max_x = this->Get_Max_Value(fourier_data_x);

		fourier_data_y = d_fourier_transform.Process(d_acceleration_data_y);
		fft_max_y = this->Get_Max_Value(fourier_data_y);

		fourier_data_z = d_fourier_transform.Process(d_acceleration_data_z);
		fft_max_z = this->Get_Max_Value(fourier_data_z);
	}

	// all graphs to the same scale
	fft_max = ZERO_EPSILON;
	
	if(fft_max_x > fft_max)
	{
		fft_max = fft_max_x;
	}

	if(fft_max_y > fft_max)
	{
		fft_max = fft_max_y;
	}
	
	if(fft_max_z > fft_max)
	{
		fft_max = fft_max_z;
	}
	
	if(d_enable_x)
	{
		this->Normalize(&fourier_data_x,fft_max);
		
		set = d_fourier_graph_widget->CreateDataSet(QColor(160,0,0));
		d_fourier_graph_widget->AddPoints(set,fourier_data_x);
	}
	
	if(d_enable_y)
	{
		this->Normalize(&fourier_data_y,fft_max);
		
		set = d_fourier_graph_widget->CreateDataSet(QColor(0,160,0));
		d_fourier_graph_widget->AddPoints(set,fourier_data_y);
	}
	
	if(d_enable_z)
	{
		this->Normalize(&fourier_data_z,fft_max);
		
		set = d_fourier_graph_widget->CreateDataSet(QColor(0,0,160));
		d_fourier_graph_widget->AddPoints(set,fourier_data_z);
	}
	
	d_fourier_graph_widget->UpdateGraph();

}

void TVibrationMeter::Update_Analysis_Data(void)
{
	double								rms;
	double								peak_peak;

	if(!d_live_display)
	{
		if(this->Calculate_Numeric_Result(d_acceleration_data_last_frame_x,&rms,&peak_peak))
		{
			d_result_x_label->setText(QString("X RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
		
		if(this->Calculate_Numeric_Result(d_acceleration_data_last_frame_y,&rms,&peak_peak))
		{
			d_result_y_label->setText(QString("Y RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
		
		if(this->Calculate_Numeric_Result(d_acceleration_data_last_frame_z,&rms,&peak_peak))
		{
			d_result_z_label->setText(QString("Z RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
	}
	else
	{
		if(this->Calculate_Numeric_Result(d_acceleration_data_x,&rms,&peak_peak))
		{
			d_result_x_label->setText(QString("X RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
		
		if(this->Calculate_Numeric_Result(d_acceleration_data_y,&rms,&peak_peak))
		{
			d_result_y_label->setText(QString("Y RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
		
		if(this->Calculate_Numeric_Result(d_acceleration_data_z,&rms,&peak_peak))
		{
			d_result_z_label->setText(QString("Z RMS: %1  P-P: %2 mm/sec^2").arg(rms,0,'f',1).arg(peak_peak,0,'f',1));
		}
	}
}

double TVibrationMeter::Get_Max_Value(
	const std::vector<TVector2>			&data) const
{
	bool								init(true);
	double								result(0.0);
	std::vector<TVector2>::const_iterator iter;
	
	for(iter = data.begin();iter != data.end();++iter)
	{
		if(init)
		{
			init = false;
			result = (*iter).y;
		}
		else
		{
			if((*iter).y > result)
			{
				result = (*iter).y;
			}
		}
	}
	
	return result;
}

void TVibrationMeter::Normalize(
	std::vector<TVector2>				* const data,
	const double						&max) const
{
	std::vector<TVector2>::iterator		iter;
	static const double					ZERO_EPSILON(0.000000001);
	
	if(max < ZERO_EPSILON)
	{
		return;
	}
	
	for(iter = data->begin();iter != data->end();++iter)
	{
		(*iter).y /= max;
	}
}

bool TVibrationMeter::Calculate_Numeric_Result(
	const std::vector<TVector2>			&data,
	double								* const rms,
	double								* const peak_peak) const
{
	double								sum_sq(0.0);
	std::vector<TVector2>::const_iterator iter;
	
	assert(rms);
	assert(peak_peak);
	
	*rms = 0.0;
	*peak_peak = 0.0;
	
	if(data.size() < 1)
	{
		return false;
	}
	
	// sum the squares and max,min
	for(iter = data.begin();iter != data.end();++iter)
	{
		sum_sq += ((*iter).y * (*iter).y);
	}
	
	// average square
	sum_sq /= static_cast<double>(data.size());
	
	*rms = sqrt(sum_sq);
	*peak_peak = (*rms) * 2.82842712474619;	// RMS to Peak-Peak conversion is 2 x sqrt(2)
	
	return true;
}



